#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
from flask import Blueprint, request
from waptserver.config import rewrite_config_item
from waptserver.decorators import requires_auth, requires_enterprise
from waptserver.app import app
from waptserver.common import make_response,make_response_from_exception

import waptserver.globalsupdate as gl
from waptserver.glpi_tasks import *

glpi = Blueprint('glpi', __name__)

@glpi.route('/api/v3/upload_to_glpi')
@requires_auth(['admin'])
@requires_enterprise
def upload_to_glpi():
    req = {}
    req['type'] = 'KO'
    if not app.conf['glpi_server_endpoint']:
        req['value'] = 'No server defined in configuration file'
        return json.dumps(req)
    if not (app.conf['glpi_server_user'] and app.conf['glpi_server_pass']):
        req['value'] = 'Please provide username and password'
        return json.dumps(req)
    upload_to_glpi_cron(force=bool(int(request.args["force"])))
    req['type'] = 'OK'
    req['value'] = 'Uploading...'
    return json.dumps(req)

@glpi.route('/api/v3/cancel_glpi_upload')
@requires_auth(['admin'])
@requires_enterprise
def cancel_glpi_upload():
    req = {}

    gl.conf['glpi_server_status'] = gl.force_get('glpi_inventory_upload_status')

    req['type'] = 'KO'
    if int(gl.conf["glpi_server_status"]) < 0 or int(gl.conf["glpi_server_status"]) == 100:
        req['value'] = 'No upload in progress'
        return json.dumps(req)
    gl.conf['glpi_server_status'] = -1
    gl.update('glpi_inventory_upload_status', gl.conf['glpi_server_status'])
    req['type'] = 'OK'
    req['value'] = 'Stopping...'
    return json.dumps(req)

def get_glpi_config():
    server_conf = {}
    values = [
        'glpi_server_endpoint',
        'glpi_server_user',
        'glpi_inventory_update_delay',
        'glpi_inventory_update_range',
        'glpi_server_pause_timeout'
        ]

    for k in values:
        if k in app.conf:
            server_conf[k] = app.conf[k]
        else:
            server_conf[k] = ''
    return server_conf

@glpi.route('/api/v3/set_glpi_server_conf', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
def set_glpi_server_conf():
    parameters = request.get_json()
    tasks_config_file = app.config['CONFIG_FILE']

    if 'glpi_server_endpoint' in parameters:
        # glpi fusion inventory url should always end with a slash
        parameters['glpi_server_endpoint'] = "%s/" % parameters['glpi_server_endpoint'].strip('/')
        app.conf['glpi_server_endpoint'] = parameters['glpi_server_endpoint']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_server_endpoint', parameters['glpi_server_endpoint'])
        app.conf['glpi_server_user'] = parameters['glpi_server_user']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_server_user', parameters['glpi_server_user'])
        app.conf['glpi_server_pass'] = parameters['glpi_server_pass']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_server_pass', parameters['glpi_server_pass'])

    if 'glpi_inventory_update_delay' in parameters:
        app.conf['glpi_inventory_update_delay'] = parameters['glpi_inventory_update_delay']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_inventory_update_delay', parameters['glpi_inventory_update_delay'])
        app.conf['glpi_inventory_update_range'] = parameters['glpi_inventory_update_range']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_inventory_update_range', parameters['glpi_inventory_update_range'])
        app.conf['glpi_server_pause_timeout'] = parameters['glpi_server_pause_timeout']
        rewrite_config_item(tasks_config_file, 'options', 'glpi_server_pause_timeout', parameters['glpi_server_pause_timeout'])

    return make_response(result = get_glpi_config(), msg='Configuration updated')

@glpi.route('/api/v3/get_glpi_server_conf', methods=['GET'])
@requires_enterprise
def get_glpi_server_conf():
    return json.dumps(get_glpi_config())

@glpi.route('/api/v3/get_glpi_upload_progress', methods=['GET'])
@requires_enterprise
def get_glpi_upload_progress():
    req = {}

    gl.conf['glpi_server_status'] = gl.force_get('glpi_inventory_upload_status')

    req['status'] = '0'
    req['progress'] = '-1'
    if (gl.conf['glpi_server_status'] == '') or (not app.conf['glpi_server_endpoint']) or (int(gl.conf['glpi_server_status']) not in range(0, 101)):
        req['status'] = '-1'
        return json.dumps(req)
    if int(gl.conf['glpi_server_status']) == 100:
        req['status'] = '1'
    req['progress'] = str(gl.conf['glpi_server_status'])
    return json.dumps(req)

@glpi.route('/api/v3/get_glpi_hosts_upload_status', methods=['GET'])
@requires_enterprise
def get_glpi_hosts_upload_status():
    req = {}

    req['type'] = 'KO'
    req['value'] = gl.force_get('glpi_hosts_upload_status')

    if not req['value']:
        return json.dumps(req)
    req['type'] = 'OK'
    return json.dumps(req)

