#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import sys
import platform
import functools
import logging
import time
import json
import traceback
import datetime

from waptserver.config import __version__
from waptserver.config import DEFAULT_CONFIG_FILE, rewrite_config_item

from waptserver.app import app

from waptutils import datetime2isodate, ensure_list, ensure_unicode, Version, setloglevel

from waptserver.utils import EWaptAuthenticationFailure, EWaptForbiddden, EWaptHostUnreachable, EWaptMissingHostData
from waptserver.utils import EWaptMissingParameter, EWaptSignalReceived, EWaptTimeoutWaitingForResult, EWaptUnknownHost
from waptserver.common import get_secured_token_generator, get_server_uuid
from waptserver.common import make_response, make_response_from_exception

from waptserver.model import wapt_db, Hosts, HostWebsocket, WaptDB

from flask import request, session
from flask_socketio import SocketIO
from flask_socketio import disconnect, send, emit


logger = logging.getLogger('waptws')

# chain SocketIO server
socketio = SocketIO(app, logger=logger, engineio_logger=logger, cors_allowed_origins='*')


def proxy_host_request(request, action):
    """Proxy a waptconsole action to wapt clients using websockets

    Args:
        uuid: can be a list or a single uuid
        notify_user: 0/1
        notify_server: 0/1

    Returns:
        dict:
            'result':
                'success' (list)
                'errors' (list)
            'msg' (str)
            'success'(bool)
            'request_time' (float)
            'error_code'
    """
    try:
        start_time = time.time()
        all_args = {k: v for k, v in request.args.items()}
        if request.json:
            all_args.update(request.json)

        uuids = ensure_list(all_args['uuid'])
        del(all_args['uuid'])
        timeout = float(request.args.get('timeout', app.conf.get('clients_read_timeout', 5)))

        result = dict(success=[], errors=[])
        tasks = []
        for uuid in uuids:
            try:
                host_data = HostWebsocket.select(HostWebsocket.host_id, HostWebsocket.server_uuid, HostWebsocket.session_id, Hosts.computer_fqdn).join(Hosts,'LEFT JOIN')\
                    .where((HostWebsocket.server_uuid == get_server_uuid()) & (HostWebsocket.host_id == uuid) & (~HostWebsocket.session_id.is_null()))\
                    .dicts()\
                    .first(1)

                if host_data:
                    msg = ''
                    logger.info(
                        'Launching %s with args %s for %s at address %s...' %
                        (action, all_args, uuid, host_data['session_id']))

                    args = dict(all_args)
                    sid = host_data['session_id']
                    computer_fqdn = host_data['computer_fqdn']

                    def emit_action(sid, uuid, action, action_args, computer_fqdn, timeout=5):
                        try:
                            got_result = []

                            def result_callback(data):
                                got_result.append(data)

                            logger.debug('Emit %s to %s (%s)' % (action, uuid, computer_fqdn))
                            socketio.emit(action, action_args, room=sid, callback=result_callback)
                            # wait for asynchronous answer...
                            wait_loop = timeout * 20
                            while not got_result:
                                wait_loop -= 1
                                if wait_loop < 0:
                                    raise EWaptTimeoutWaitingForResult('Timeout, client did not send result within %s s' % timeout)
                                socketio.sleep(0.05)

                            # action succedded
                            result['success'].append(
                                dict(
                                    uuid=uuid,
                                    msg=msg,
                                    computer_fqdn=computer_fqdn,
                                    result=got_result[0],
                                ))
                        except Exception as e:
                            result['errors'].append(
                                dict(
                                    uuid=uuid,
                                    msg='%s' % repr(e),
                                    computer_fqdn='',
                                ))
                    if sid:
                        tasks.append(socketio.start_background_task(
                            emit_action, sid=sid, uuid=uuid, action=action, action_args=args, computer_fqdn=computer_fqdn))

                else:
                    result['errors'].append(
                        dict(
                            uuid=uuid,
                            msg='Host %s is not registered or not connected wia websockets' % uuid,
                            computer_fqdn='',
                        ))
            except Exception as e:
                result['errors'].append(
                    dict(
                        uuid=uuid,
                        msg='Host %s error %s' % (uuid, repr(e)),
                        computer_fqdn='',
                    ))

        # wait for all background tasks to terminate or timeout...
        # assume the timeout should be longer if more hosts to perform
        wait_loop = timeout * len(tasks)
        while True:
            running = [t for t in tasks if t.is_alive()]
            if not running:
                break
            wait_loop -= 1
            if wait_loop < 0:
                break
            socketio.sleep(0.05)

        msg = ['Success : %s, Errors: %s' % (len(result['success']), len(result['errors']))]
        if result['errors']:
            msg.extend(['%s: %s' % (e['computer_fqdn'], e['msg'])
                        for e in result['errors']])

        return make_response(result,
                             msg='\n- '.join(msg),
                             success=len(result['success']) > 0,
                             request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


# SocketIO Callbacks / handlers
@socketio.on('trigger_update_result')
def on_trigger_update_result(result):
    """Return from update on client"""
    logger.debug('Trigger Update result : %s (uuid:%s)' % (result, request.args['uuid']))
    # send to all waptconsole warching this host.
    socketio.emit('trigger_update_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_upgrade_result')
def on_trigger_upgrade_result(result):
    """Return from the launch of upgrade on a client"""
    logger.debug('Trigger Upgrade result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_upgrade_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_install_packages_result')
def on_trigger_install_packages_result(result):
    logger.debug('Trigger install result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_install_packages_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_remove_packages_result')
def on_trigger_remove_packages_result(result):
    logger.debug('Trigger remove result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_remove_packages_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('reconnect')
@socketio.on('connect')
def on_waptclient_connect():
    try:
        uuid = request.args.get('uuid', None)
        if not uuid:
            uuid = request.headers.get('uuid', None)
        if not uuid:
            raise EWaptForbiddden('Missing source host uuid')

        allow_unauthenticated_connect = app.conf.get('allow_unauthenticated_connect', False)
        if not allow_unauthenticated_connect:
            try:
                client_token = request.args.get('token',None)
                if not client_token:
                    authorization = request.headers.get('authorization')
                    if authorization:
                        (bearer,client_token) = authorization.split(' ',1)
                        if bearer != 'Bearer' or not client_token:
                            raise EWaptForbiddden('Bad header authorization token')

                if not client_token:
                    raise EWaptAuthenticationFailure('No authorization token')

                token_gen = get_secured_token_generator()
                token_data = token_gen.loads(client_token)
                uuid = token_data.get('uuid', None)
                if not uuid:
                    raise EWaptAuthenticationFailure('Bad host UUID')
                if token_data['server_uuid'] != get_server_uuid():
                    raise EWaptAuthenticationFailure('Bad server UUID')
            except Exception as e:
                raise EWaptAuthenticationFailure('SocketIO connection not authorized, invalid token: %s' % e)
            logger.info('Socket.IO connection from wapt client sid %s (uuid: %s fqdn:%s)' % (request.sid, uuid, token_data.get('computer_fqdn')))
        else:
            logger.info('Unauthenticated Socket.IO connection from wapt client sid %s (uuid: %s)' % (request.sid, uuid))

        with WaptDB():
            # update the db
            with wapt_db.atomic() as trans:
                sid = HostWebsocket.select(HostWebsocket.session_id).where(HostWebsocket.host_id == uuid).tuples().first()
                # this record is initialized (inserted) with get_websocket_auth_token, sid is initialized to None
                # if this host is not registered, or sid in database is not the same as current connection, reject the connection
                # protect against one client with several established websockets (bug on client side)
                if not sid or (sid[0] and sid[0] != request.sid):
                    return False

                # new connection
                if not sid[0]:
                    # stores new sid in database
                    HostWebsocket.update(
                        server_uuid=get_server_uuid(),
                        session_id=request.sid,
                        timestamp=datetime.datetime.now(),
                    ).where(HostWebsocket.host_id == uuid).execute()
            session['uuid'] = uuid
        return True

    except Exception as e:
        if 'uuid' in session:
            session.pop('uuid')
        logger.warning('SocketIO connection refused for uuid %s, sid %s: %s, instance %s' % (uuid, request.sid, e, app.conf.get('application_root')))
        return False


@socketio.on('wapt_pong')
def on_wapt_pong():
    uuid = None
    try:
        uuid = session.get('uuid')
        if not uuid:
            logger.critical('SocketIO %s connected but no host uuid in session: asking connected host to reconnect' % (request.sid))
            emit('wapt_force_reconnect')
            return False
        else:
            logger.debug('Socket.IO pong from wapt client sid %s (uuid: %s)' % (request.sid, session.get('uuid', None)))
            # stores sid in database

        with WaptDB():
            # update the db
            with wapt_db.atomic() as trans:
                sid = HostWebsocket.select(HostWebsocket.session_id).where(HostWebsocket.host_id == uuid).tuples().first()
                # this record is initialized (inserted) with get_websocket_auth_token, sid is initialized to None
                # if this host is not registered, or sid is not the same as current, reject the connection
                # protect against one client with several established websockets (bug on client side)
                if not sid or (sid[0] and sid[0] != request.sid):
                    return False

                # stores new sid in database
                HostWebsocket.update(
                    server_uuid=get_server_uuid(),
                    session_id=request.sid,
                    timestamp=datetime.datetime.now(),
                ).where(HostWebsocket.host_id == uuid).execute()

            session['uuid'] = uuid
        return True

    except Exception as e:
        logger.critical('SocketIO pong error for uuid %s and sid %s : %s, instance: %s' % (uuid, request.sid, traceback.format_exc(), app.conf.get('application_root')))
        return False


@socketio.on('disconnect')
def on_waptclient_disconnect():
    uuid = session.get('uuid', None)
    logger.info('Socket.IO disconnection from wapt client sid %s (uuid: %s)' % (request.sid, uuid))
    # clear sid in database
    with WaptDB():
        with wapt_db.atomic() as trans:
            # stores new sid in database
            res = HostWebsocket.update(
                session_id=None,
                timestamp=datetime.datetime.now(),
            ).where((HostWebsocket.host_id == uuid)).execute()
            logger.debug('Set websocket sid to none in database for UUID %s : count=%s' % (uuid,res))
    return True

@socketio.on_error()
def on_wapt_socketio_error(e):
    logger.critical('Socket IO : An error has occurred for sid %s, uuid:%s : %s' % (request.sid, request.args.get('uuid', None), repr(e)))

# end websockets
